==> klayout-0.19.3.tar.gz: klayout-0.19.3/src/contrib/gds2_txt/dbGDS2TextReader.cc <==

/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbGDS2Reader.h"
#include "dbGDS2Converter.h"
#include "dbGDS2.h"
#include "dbGDS2TextReader.h"
#include "dbArray.h"

#include "tlException.h"
#include "tlString.h"
#include "tlClassRegistry.h"

namespace db
{

GDS2ReaderText::GDS2ReaderText(tl::InputStream &s, int _iChunkSize)
  : GDS2ReaderBase(), pXYData(0), uXYLength(0), sStream(s),
    mProgress (tl::translate ("Reading GDS2 text file"), 0, 10000)
{
  mProgress.set_format (tl::translate ("%.0f MB"));
  mProgress.set_res (1.0);
  mProgress.set_unit (1024 * 1024);

  vInit(_iChunkSize);
}

GDS2ReaderText::~GDS2ReaderText()
{
  if (pXYData) {
    delete[] pXYData;
    pXYData = 0;
  }
}
  
const LayerMap &
GDS2ReaderText::read (db::Layout &layout, const db::LoadLayoutOptions &options) throw (tl::Exception)
{
  //  HINT: reuse the standard GDS2 reader options for the text reader. 
  //  However, the allow_big_records and allow_multi_xy_records options are ignored.
  db::GDS2ReaderOptions gds2_options;

  const db::GDS2ReaderOptions *specific_options = dynamic_cast<const db::GDS2ReaderOptions *> (options.get_specific_options ("GDS2"));
  if (specific_options) {
    gds2_options = *specific_options;
  }

  m_iNumLine = 0;

  return basic_read (layout, options, false, gds2_options.box_mode);
}

const LayerMap &
GDS2ReaderText::read (db::Layout &layout) throw (tl::Exception)
{
  return read (layout, db::LoadLayoutOptions ());
}

short 
GDS2ReaderText::get_record () throw (tl::Exception)
{
   short siValueToReturn   = 0;
   std::string sLocalArg    = "";
   sExtractedArguments    = "";

   std::string sLine = "";
   
   do
   {   
     sLine.clear ();

     if(sExtractedValue == "")
     {
       const char *next = sStream.get (1);
       if (! next) {
         break; // at EOF
       }

       while (next != 0 && *next != '\n') {
         sLine += *next;
         next = sStream.get (1);
       }

       ++ m_iNumLine;
       
        //remove comments
        if(sLine != "" && sLine.at(0) == '#')
        {
          sLine = "";
        }

        //remove unix-like characters
        if(sLine != "" && sLine.at(sLine.size() -1) == '\r')
        {
          sLine = sLine.substr(0, sLine.size() -1);
        }
 
        //remove useless spaces
        while(sLine != "" && sLine.at(0) == ' ')
        {
          sLine = sLine.substr(1);
        }
        //append data to the currently read values
        sExtractedValue.append(sLine); 
      }

      // Save data, so that they can be re-evaluated later, as for now on we might only look for the end of the element
      sNextLine = sExtractedValue;

      if(sExtractedValue != "")
      {
        std::string s1 = "", sNewArgument = "";
        short siLocalvalue = siExtactData(sExtractedValue, s1, sNewArgument);
        
        // In the function below, we treat  sXY identifier in some different manner, in order to gain some speed, as it require special actions

        if(siLocalvalue) // If the extracted data does contain an identifier
        {
          if(!siValueToReturn) // If we do not have stored an identifier
          {
            siValueToReturn = siLocalvalue;
                                
            if(siValueToReturn == sXY)
            {
                vConvertToXY(sNewArgument);    
             }
            else // save extracted arguments
            {
              sExtractedArguments.append(" ");
              sExtractedArguments.append(sNewArgument);
            }

            if(siLocalvalue == sENDLIB) // Special case for the end of librarie detection
            {
              vInit(iChunkSize);
              return siLocalvalue;
            }
          }
          else
          {
            while(sExtractedArguments != "" && sExtractedArguments.at(0) == ' ')
            {
              sExtractedArguments = sExtractedArguments.substr(1);
            }

            if(siValueToReturn == sXY)
            {
              vConvertToXY(sNewArgument);              
              vConvertToStruct();
            }
            sExtractedValue = sNextLine;
            return  siValueToReturn;
          }
        }
        else
        {
            if((int)ssFormattingStream.tellg() == (int) sStream.pos())
            {
              return siLocalvalue;
            }

            if(siValueToReturn == sXY)
            {
              vConvertToXY(sNewArgument);
            }
        }
      }
  }while(sExtractedValue == "");

  //We should NEVER reach this point. Sending an error
  error(tl::translate("Unhandled error occured during stream read"));
  return 0;
}



void 
GDS2ReaderText::vConvertToStruct()
{
  //memory management
  if(pXYData)
  {
    delete[] pXYData;
  }


  pXYData = new GDS2XY[aulPoints.size()];

  uXYLength = aulPoints.size();

  int iCharIndex = 0;
  unsigned char *pChar = (unsigned char *) pXYData;

  // Save all the point of the XY delimiter, so that it can be loaded in the same way as the basic
  // GDS Binary reader. This piece of code is quite ugly, but function calls are reduced to the minimum
  // in order to increase speed, as the structure might be quite big if we deal with large files.
  for(unsigned int uiVectorIndex = 0; uiVectorIndex < aulPoints.size(); ++ uiVectorIndex)
  {
    std::pair<long, long> ululXYPair = aulPoints.at(uiVectorIndex);

    pChar[iCharIndex]   = (unsigned char)floor(ululXYPair.first / 16777216.0);
    ululXYPair.first    -= pChar[iCharIndex++] * 16777216;
    pChar[iCharIndex]   = (unsigned char)floor(ululXYPair.first / 65536.0);
    ululXYPair.first    -= pChar[iCharIndex++] * 65536;
    pChar[iCharIndex]   = (unsigned char)floor(ululXYPair.first / 256.0);
    ululXYPair.first    -= pChar[iCharIndex++] * 256;
    pChar[iCharIndex++] = (unsigned char)ululXYPair.first;
    
    pChar[iCharIndex]   = (unsigned char)floor(ululXYPair.second / 16777216.0);
    ululXYPair.second   -= pChar[iCharIndex++] * 16777216;
    pChar[iCharIndex]   = (unsigned char)floor(ululXYPair.second / 65536.0);
    ululXYPair.second   -= pChar[iCharIndex++] * 65536;
    pChar[iCharIndex]   = (unsigned char)floor(ululXYPair.second / 256.0);
    ululXYPair.second   -= pChar[iCharIndex++] * 256;
    pChar[iCharIndex++] = (unsigned char)ululXYPair.second; 
  }
  aulPoints.clear();
}

void 
GDS2ReaderText::vConvertToXY(std::string &_sArg)
{
  unsigned int uiDelimPos = _sArg.find(":");
  
  if(uiDelimPos != _sArg.npos)
  {
    std::string sX = "", sY = "";

    sX = _sArg.substr(0, uiDelimPos);
    sY = _sArg.substr(uiDelimPos + 1);

    //data may be read outside of the allocated memory, so we must keep them in the extracted arguments too
    sExtractedArguments.append(" ");
    sExtractedArguments.append(sX);
    sExtractedArguments.append(" ");
    sExtractedArguments.append(sY);

    long ulX, ulY;
    ulX = (long) atof(sX.c_str());
    ulY = (long) atof(sY.c_str());
    aulPoints.push_back(std::pair<long, long>(ulX, ulY));
  }
  _sArg = "";
}

short 
GDS2ReaderText::siExtactData(std::string &_sInput, std::string &_sToken, std::string &_sArguments)
{
  if(_sInput != "")
  {
    std::string sInformation = "", sLocalArguments = "";
    
    // Getting rid of the ";" character
    if(_sInput.find(";") != _sInput.npos)
    {
      sInformation = _sInput.substr(0, _sInput.find(";"));
      
      if( _sInput.size() > sInformation.size() + 2)
      {
          _sInput = _sInput.substr(_sInput.find(";") + 1);
      }
      else
      {
        _sInput = "";
      }
    }
    else
    {
      sInformation    = _sInput;
      _sInput         = "";
    }

    //No space => No  arguments
    if(sInformation.find(" ") == sInformation.npos)
    {
      _sToken         = sInformation;
      sLocalArguments = "";
    }
    else
    {
      _sToken         = sInformation.substr( 0, sInformation.find(" "));
      sLocalArguments = sInformation.substr(_sToken.size());
    }

    while(_sToken != "" && _sToken.at(0) == ' ')
    {
      _sToken = _sToken.substr(1);
    }

    while(sLocalArguments != "" && sLocalArguments.at(0) == ' ')
    {
      sLocalArguments = sLocalArguments.substr(1);
    }

    short siTokenConversion = db::gds2_converter.to_short(_sToken.c_str());

    if(siTokenConversion)
    {
      _sArguments.append(" ");
      _sArguments.append(sLocalArguments);
      return siTokenConversion;
    }
    _sToken.append(" ");
    _sToken.append(sLocalArguments);
    _sToken.append(" ");
    _sArguments.append(_sToken);
    _sToken = "";
  }
  return 0;
}

const char *
GDS2ReaderText::get_string () throw (tl::Exception)
{
  return sExtractedArguments.c_str();
}

double
GDS2ReaderText::get_double () throw (tl::Exception)
{
  std::string sContainer = "";

  while(sExtractedArguments != "" && sExtractedArguments.at(0) == ' ')
  {
    sExtractedArguments = sExtractedArguments.substr(1);
  }

  if(sExtractedArguments.find(" ") != sExtractedArguments.npos)
  {
    sContainer = sExtractedArguments.substr(0, sExtractedArguments.find(" "));
    sExtractedArguments = sExtractedArguments.substr(sContainer.size() + 1);
  }
  else if (sExtractedArguments.find(";") != sExtractedArguments.npos)
  {
    sContainer = sExtractedArguments.substr(0, sExtractedArguments.find(";"));
    sExtractedArguments = sExtractedArguments.substr(sContainer.size() + 1);
  }
  else
  {
    sContainer = sExtractedArguments;
    sExtractedArguments = "";
  }
  return atof(sContainer.c_str());
}

void
GDS2ReaderText::get_string (tl::string &s) const throw (tl::Exception)
{
  s.assign(sExtractedArguments.c_str(), 0, sExtractedArguments.size());
}

int 
GDS2ReaderText::get_int () throw (tl::Exception)
{
   return (int)get_double ();
}

short 
GDS2ReaderText::get_short () throw (tl::Exception)
{
   return (short) get_double ();
}

unsigned short 
GDS2ReaderText::get_ushort () throw (tl::Exception)
{
   return (unsigned short) get_double ();
}

void 
GDS2ReaderText::error (const std::string &msg) throw (tl::Exception)
{
  throw GDS2ReaderTextException (msg,  m_iNumLine, cellname().c_str ());
}

void 
GDS2ReaderText::warn (const std::string &msg) 
{
  // TODO: this is simple handling ..
  fprintf (stderr, tl::translate ("WARNING: %s (line number=%ld, cell=%s)\n").c_str (), msg.c_str (), m_iNumLine, cellname().c_str ());
}

short  
GDS2ReaderText::peek_record () throw (tl::Exception)
{
  return 0; // is used to detect multiple XY records only - does not matter what it returns.
}

GDS2XY * 
GDS2ReaderText::get_xy_data (unsigned int &xy_length) throw (tl::Exception)
{
  xy_length = uXYLength;
  return pXYData;
}

void 
GDS2ReaderText::progress_checkpoint ()
{
  mProgress.set (sStream.pos ());
}

void 
GDS2ReaderText::vInit(int _iChunkSize)
{

  if(pXYData)
  {
    delete[] pXYData;
    pXYData = 0;
  }
  m_iNumLine           = 0;
  sExtractedValue      = ""; 
  sExtractedArguments  = ""; 
  sNextLine            = "";
  iGetSize             = _iChunkSize;
  iChunkSize           = _iChunkSize;
  iNumRead             = 0;
  ssFormattingStream.clear();
  ssFormattingStream.str("");
  ssFormattingStream.seekg(0, std::ios::beg);
}


}// end namespace db

==> klayout-0.21.3.tar.gz: klayout-0.21.3/src/contrib/gds2_txt/dbGDS2TextReader.cc <==

/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbGDS2Reader.h"
#include "dbGDS2Converter.h"
#include "dbGDS2.h"
#include "dbGDS2TextReader.h"
#include "dbArray.h"

#include "tlException.h"
#include "tlString.h"
#include "tlClassRegistry.h"

namespace db
{

GDS2ReaderText::GDS2ReaderText(tl::InputStreamBase &s, int _iChunkSize)
  : GDS2ReaderBase(), pXYData(0), uXYLength(0), sStream(s),
    mProgress (tl::translate ("Reading GDS2 text file"), 0, 10000)
{
  mProgress.set_format (tl::translate ("%.0f MB"));
  mProgress.set_res (1.0);
  mProgress.set_unit (1024 * 1024);

  vInit(_iChunkSize);
}

GDS2ReaderText::~GDS2ReaderText()
{
  if (pXYData) {
    delete[] pXYData;
    pXYData = 0;
  }
}
  
const LayerMap &
GDS2ReaderText::read (db::Layout &layout, const db::LoadLayoutOptions &options) throw (tl::Exception)
{
  //  HINT: reuse the standard GDS2 reader options for the text reader. 
  //  However, the allow_big_records and allow_multi_xy_records options are ignored.
  db::GDS2ReaderOptions gds2_options;

  const db::GDS2ReaderOptions *specific_options = dynamic_cast<const db::GDS2ReaderOptions *> (options.get_specific_options ("GDS2"));
  if (specific_options) {
    gds2_options = *specific_options;
  }

  m_iNumLine = 0;

  return basic_read (layout, options, false, gds2_options.box_mode);
}

const LayerMap &
GDS2ReaderText::read (db::Layout &layout) throw (tl::Exception)
{
  return read (layout, db::LoadLayoutOptions ());
}

short 
GDS2ReaderText::get_record () throw (tl::Exception)
{
   short siValueToReturn   = 0;
   std::string sLocalArg    = "";
   sExtractedArguments    = "";

   std::string sLine = "";
   
   do
   {   
     sLine.clear ();

     if(sExtractedValue == "")
     {
       const char *next = sStream.get (1);
       if (! next) {
         break; // at EOF
       }

       while (next != 0 && *next != '\n') {
         sLine += *next;
         next = sStream.get (1);
       }

       ++ m_iNumLine;
       
        //remove comments
        if(sLine != "" && sLine.at(0) == '#')
        {
          sLine = "";
        }

        //remove unix-like characters
        if(sLine != "" && sLine.at(sLine.size() -1) == '\r')
        {
          sLine = sLine.substr(0, sLine.size() -1);
        }
 
        //remove useless spaces
        while(sLine != "" && sLine.at(0) == ' ')
        {
          sLine = sLine.substr(1);
        }
        //append data to the currently read values
        sExtractedValue.append(sLine); 
      }

      // Save data, so that they can be re-evaluated later, as for now on we might only look for the end of the element
      sNextLine = sExtractedValue;

      if(sExtractedValue != "")
      {
        std::string s1 = "", sNewArgument = "";
        short siLocalvalue = siExtactData(sExtractedValue, s1, sNewArgument);
        
        // In the function below, we treat  sXY identifier in some different manner, in order to gain some speed, as it require special actions

        if(siLocalvalue) // If the extracted data does contain an identifier
        {
          if(!siValueToReturn) // If we do not have stored an identifier
          {
            siValueToReturn = siLocalvalue;
                                
            if(siValueToReturn == sXY)
            {
                vConvertToXY(sNewArgument);    
             }
            else // save extracted arguments
            {
              sExtractedArguments.append(" ");
              sExtractedArguments.append(sNewArgument);
            }

            if(siLocalvalue == sENDLIB) // Special case for the end of librarie detection
            {
              vInit(iChunkSize);
              return siLocalvalue;
            }
          }
          else
          {
            while(sExtractedArguments != "" && sExtractedArguments.at(0) == ' ')
            {
              sExtractedArguments = sExtractedArguments.substr(1);
            }

            if(siValueToReturn == sXY)
            {
              vConvertToXY(sNewArgument);              
              vConvertToStruct();
            }
            sExtractedValue = sNextLine;
            return  siValueToReturn;
          }
        }
        else
        {
            if((int)ssFormattingStream.tellg() == (int) sStream.pos())
            {
              return siLocalvalue;
            }

            if(siValueToReturn == sXY)
            {
              vConvertToXY(sNewArgument);
            }
        }
      }
  }while(sExtractedValue == "");

  //We should NEVER reach this point. Sending an error
  error(tl::translate("Unhandled error occured during stream read"));
  return 0;
}



void 
GDS2ReaderText::vConvertToStruct()
{
  //memory management
  if(pXYData)
  {
    delete[] pXYData;
  }


  pXYData = new GDS2XY[aulPoints.size()];

  uXYLength = (unsigned int) aulPoints.size();

  int iCharIndex = 0;
  unsigned char *pChar = (unsigned char *) pXYData;

  // Save all the point of the XY delimiter, so that it can be loaded in the same way as the basic
  // GDS Binary reader. This piece of code is quite ugly, but function calls are reduced to the minimum
  // in order to increase speed, as the structure might be quite big if we deal with large files.
  for(unsigned int uiVectorIndex = 0; uiVectorIndex < aulPoints.size(); ++ uiVectorIndex)
  {
    std::pair<long, long> ululXYPair = aulPoints.at(uiVectorIndex);

    pChar[iCharIndex++] = (ululXYPair.first >> 24);
    pChar[iCharIndex++] = (ululXYPair.first >> 16);
    pChar[iCharIndex++] = (ululXYPair.first >> 8);
    pChar[iCharIndex++] = ululXYPair.first;
    
    pChar[iCharIndex++] = (ululXYPair.second >> 24);
    pChar[iCharIndex++] = (ululXYPair.second >> 16);
    pChar[iCharIndex++] = (ululXYPair.second >> 8);
    pChar[iCharIndex++] = ululXYPair.second;
  }
  aulPoints.clear();
}

void 
GDS2ReaderText::vConvertToXY(std::string &_sArg)
{
  size_t uiDelimPos = _sArg.find(":");
  
  if(uiDelimPos != _sArg.npos)
  {
    std::string sX = "", sY = "";

    sX = _sArg.substr(0, uiDelimPos);
    sY = _sArg.substr(uiDelimPos + 1);

    //data may be read outside of the allocated memory, so we must keep them in the extracted arguments too
    sExtractedArguments.append(" ");
    sExtractedArguments.append(sX);
    sExtractedArguments.append(" ");
    sExtractedArguments.append(sY);

    long ulX, ulY;
    ulX = (long) atof(sX.c_str());
    ulY = (long) atof(sY.c_str());
    aulPoints.push_back(std::pair<long, long>(ulX, ulY));
  }
  _sArg = "";
}

short 
GDS2ReaderText::siExtactData(std::string &_sInput, std::string &_sToken, std::string &_sArguments)
{
  if(_sInput != "")
  {
    std::string sInformation = "", sLocalArguments = "";
    
    // Getting rid of the ";" character
    if(_sInput.find(";") != _sInput.npos)
    {
      sInformation = _sInput.substr(0, _sInput.find(";"));
      
      if( _sInput.size() > sInformation.size() + 2)
      {
          _sInput = _sInput.substr(_sInput.find(";") + 1);
      }
      else
      {
        _sInput = "";
      }
    }
    else
    {
      sInformation    = _sInput;
      _sInput         = "";
    }

    //No space => No  arguments
    if(sInformation.find(" ") == sInformation.npos)
    {
      _sToken         = sInformation;
      sLocalArguments = "";
    }
    else
    {
      _sToken         = sInformation.substr( 0, sInformation.find(" "));
      sLocalArguments = sInformation.substr(_sToken.size());
    }

    while(_sToken != "" && _sToken.at(0) == ' ')
    {
      _sToken = _sToken.substr(1);
    }

    while(sLocalArguments != "" && sLocalArguments.at(0) == ' ')
    {
      sLocalArguments = sLocalArguments.substr(1);
    }

    short siTokenConversion = db::gds2_converter.to_short(_sToken.c_str());

    if(siTokenConversion)
    {
      _sArguments.append(" ");
      _sArguments.append(sLocalArguments);
      return siTokenConversion;
    }
    _sToken.append(" ");
    _sToken.append(sLocalArguments);
    _sToken.append(" ");
    _sArguments.append(_sToken);
    _sToken = "";
  }
  return 0;
}

const char *
GDS2ReaderText::get_string () throw (tl::Exception)
{
  return sExtractedArguments.c_str();
}

double
GDS2ReaderText::get_double () throw (tl::Exception)
{
  std::string sContainer = "";

  while(sExtractedArguments != "" && sExtractedArguments.at(0) == ' ')
  {
    sExtractedArguments = sExtractedArguments.substr(1);
  }

  if(sExtractedArguments.find(" ") != sExtractedArguments.npos)
  {
    sContainer = sExtractedArguments.substr(0, sExtractedArguments.find(" "));
    sExtractedArguments = sExtractedArguments.substr(sContainer.size() + 1);
  }
  else if (sExtractedArguments.find(";") != sExtractedArguments.npos)
  {
    sContainer = sExtractedArguments.substr(0, sExtractedArguments.find(";"));
    sExtractedArguments = sExtractedArguments.substr(sContainer.size() + 1);
  }
  else
  {
    sContainer = sExtractedArguments;
    sExtractedArguments = "";
  }
  return atof(sContainer.c_str());
}

void
GDS2ReaderText::get_string (tl::string &s) const throw (tl::Exception)
{
  s.assign(sExtractedArguments.c_str(), 0, sExtractedArguments.size());
}

int 
GDS2ReaderText::get_int () throw (tl::Exception)
{
   return (int)get_double ();
}

short 
GDS2ReaderText::get_short () throw (tl::Exception)
{
   return (short) get_double ();
}

unsigned short 
GDS2ReaderText::get_ushort () throw (tl::Exception)
{
   return (unsigned short) get_double ();
}

void 
GDS2ReaderText::error (const std::string &msg) throw (tl::Exception)
{
  throw GDS2ReaderTextException (msg,  m_iNumLine, cellname().c_str ());
}

void 
GDS2ReaderText::warn (const std::string &msg) 
{
  // TODO: this is simple handling ..
  fprintf (stderr, tl::translate ("WARNING: %s (line number=%ld, cell=%s)\n").c_str (), msg.c_str (), m_iNumLine, cellname().c_str ());
}

short  
GDS2ReaderText::peek_record () throw (tl::Exception)
{
  return 0; // is used to detect multiple XY records only - does not matter what it returns.
}

GDS2XY * 
GDS2ReaderText::get_xy_data (unsigned int &xy_length) throw (tl::Exception)
{
  xy_length = uXYLength;
  return pXYData;
}

void 
GDS2ReaderText::progress_checkpoint ()
{
  mProgress.set (sStream.pos ());
}

void 
GDS2ReaderText::vInit(int _iChunkSize)
{

  if(pXYData)
  {
    delete[] pXYData;
    pXYData = 0;
  }
  m_iNumLine           = 0;
  sExtractedValue      = ""; 
  sExtractedArguments  = ""; 
  sNextLine            = "";
  iGetSize             = _iChunkSize;
  iChunkSize           = _iChunkSize;
  iNumRead             = 0;
  ssFormattingStream.clear();
  ssFormattingStream.str("");
  ssFormattingStream.seekg(0, std::ios::beg);
}


}// end namespace db
